package com.gisdo.opengis.Utils;

import com.gisdo.opengis.Model.GdalResult;
import org.apache.poi.ss.usermodel.*;
import org.apache.poi.ss.util.CellRangeAddress;
import org.apache.poi.xssf.usermodel.XSSFColor;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.gdal.gdal.gdal;
import org.gdal.ogr.Feature;
import org.gdal.ogr.ogr;
import org.gdal.osr.SpatialReference;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Path;
import java.text.DecimalFormat;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

public  class GdalUtils {
    public static void InitGdal(){
        //注册gdal驱动
        gdal.AllRegister();
        //注册ogr驱动
        ogr.RegisterAll();
        //设置支持中文
		gdal.SetConfigOption("GDAL_FILENAME_IS_UTF8","YES");
		gdal.SetConfigOption("SHAPE_ENCODING","CP936");
    }
    static{
        InitGdal();
    }
    /***
     * 获取图层数量
     * @param path 数据路径
     * @return 返回图层数量
     */
    public static GdalResult<Integer> GetLayerCount(String path){
        GdalResult<Integer> result  ;
        var dataset = ogr.Open(path,0);
        if(dataset == null){
            return new  GdalResult<Integer>("打开数据源失败",0);
        }else{
            var layerCount = dataset.GetLayerCount();
            dataset.delete();
            System.err.println("数据源打开成功,图层数量"+layerCount);
            return new GdalResult<Integer>("数据源打开成功,图层数量",layerCount);
        }
    }

    /***
     * 获取图层数量
     * @param path 数据路径
     * @return 获取返回要素数量
     */
    public static GdalResult<Long> GetFeatureCount(String path) {
        var dataset = ogr.Open(path, 0);
        if (dataset == null) {
            return new GdalResult<>("打开数据源失败", -1L);
        } else {
            var layer = dataset.GetLayer(0);
            if (layer == null) {
                dataset.delete();
                return new GdalResult<>("无法打开图层", -1L);
            }
            var featureCount = layer.GetFeatureCount();
            System.err.println("数据源打开成功,要素数量" + featureCount + ",图层名称" + layer.GetName());
            dataset.delete();
            return new GdalResult<>("要素数量获取成功", featureCount);
        }
    }

    /***
     * 获取图层面积
     * @param path 图层路径
     * @return 获取投影面积（若无投影坐标则默认使用3857投影坐标系）
     */
    public static GdalResult<Double> GetSummaryArea(String path) {
        var dataset = ogr.Open(path, 0);
        double totalArea = 0.0;
        if (dataset == null) {
            System.err.println("无法打开数据源" + path);
            return new GdalResult<>("无法打开数据源", -1.0);
        }
        var layer = dataset.GetLayer(0);
        if (layer == null) {
            dataset.delete();
            System.err.println("无法打开图层");
            return new GdalResult<>("无法打开图层", -1.0);
        }
        var spatialRef = layer.GetSpatialRef();
        if (spatialRef == null) {
            dataset.delete();
            System.err.println("无法获取坐标系");
            return new GdalResult<>("无法获取坐标系", -1.0);
        }
        int isProjected = spatialRef.IsProjected();
        SpatialReference targetSRS = new SpatialReference();
        targetSRS.ImportFromEPSG(3857); // 默认转换为 EPSG:3857
        Feature feature;
        while ((feature = layer.GetNextFeature()) != null) {
            var geometry = feature.GetGeometryRef();
            if (geometry != null) {
                // 如果未投影，转换为 EPSG:3857
                if (isProjected == 0) {
                    geometry.TransformTo(targetSRS);
                }
                totalArea += geometry.Area();
            }
            feature.delete();
        }

        dataset.delete();
        return new GdalResult<>("图层总面积计算成功", totalArea/1000000.0);
    }

    /***
     * 获取要素面积
     * @param path
     * @return
     */
    public static GdalResult<Map<String,Double>> GetFeatureArea(String path) {
        var dataset = ogr.Open(path, 0);
        double totalArea = 0.0;
        if (dataset == null) {
            System.err.println("无法打开数据源" + path);
            return new GdalResult<>("无法打开数据源", null);
        }
        var layer = dataset.GetLayer(0);
        if (layer == null) {
            dataset.delete();
            System.err.println("无法打开图层");
            return new GdalResult<>("无法打开图层", null);
        }
        var spatialRef = layer.GetSpatialRef();
        if (spatialRef == null) {
            dataset.delete();
            System.err.println("无法获取坐标系");
            return new GdalResult<>("无法获取坐标系", null);
        }
        int isProjected = spatialRef.IsProjected();
        SpatialReference targetSRS = new SpatialReference();
        targetSRS.ImportFromEPSG(3857); // 默认转换为 EPSG:3857
        var areaMap = new HashMap<String, Double>();
        Feature feature;
        while ((feature = layer.GetNextFeature()) != null) {
            var geometry = feature.GetGeometryRef();
            if (geometry != null) {
                // 如果未投影，转换为 EPSG:3857
                if (isProjected == 0) {
                    geometry.TransformTo(targetSRS);
                }
                var featureArea = geometry.GetArea()/1000000.0;
                totalArea+=featureArea;
                areaMap.put(feature.GetFieldAsString(0),featureArea);
            }
            areaMap.put("total",totalArea);
            feature.delete();
        }

        dataset.delete();
        return new GdalResult<>("图层要素面积计算成功", areaMap);
    }

    /***
     * 面积制表
     * @param path
     * @throws IOException
     */
    public static void ExportAreaAsTable(String path) throws IOException {
        var dataset = ogr.Open(path, 0);
        if (dataset == null) {
            System.err.println("打开数据源失败");
            return;
        }

        var layer = dataset.GetLayer(0);
        if (layer == null) {
            dataset.delete();
            System.err.println("无法打开图层");
            return;
        }

        // 确定保存文件路径
        String basePath = path.split("\\.")[0];
        String pathToSave = basePath + ".xlsx";
        File outputFile = new File(pathToSave);
        var formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd-HH-mm-ss");

        // 如果文件已存在，则重新命名
        if (outputFile.exists()) {
            var time = formatter.format(LocalDateTime.now());
            String newFileName = basePath+ "-"+ time   + ".xlsx";
            pathToSave = newFileName;
        }

        // 创建 Excel 工作簿
        Workbook workbook = new XSSFWorkbook();
        Sheet sheet = workbook.createSheet(layer.GetName());

        // 创建表头样式
        CellStyle headerStyle = workbook.createCellStyle();
        headerStyle.setFillForegroundColor(IndexedColors.GREY_25_PERCENT.getIndex());
        headerStyle.setFillPattern(FillPatternType.SOLID_FOREGROUND);
        headerStyle.setAlignment(HorizontalAlignment.CENTER);
        headerStyle.setBorderBottom(BorderStyle.THIN);
        headerStyle.setBorderLeft(BorderStyle.THIN);
        headerStyle.setBorderRight(BorderStyle.THIN);
        headerStyle.setBorderTop(BorderStyle.THIN);
        // 创建表头行
        Row headerRow = sheet.createRow(0);
        var layerDefn = layer.GetLayerDefn();
        int fieldCount = layerDefn.GetFieldCount();
        // 写入字段名称作为表头
        for (int i = 0; i < fieldCount; i++) {
            String fieldName = layerDefn.GetFieldDefn(i).GetName();
            Cell cell = headerRow.createCell(i);
            cell.setCellValue(fieldName);
            cell.setCellStyle(headerStyle);
        }
        headerRow.createCell(fieldCount).setCellStyle(headerStyle);
        headerRow.getCell(fieldCount).setCellValue("要素面积(平方公里)");
        // 遍历图层的要素，写入数据
        int rowIndex = 1;
        Feature feature;
        var resultArea =  GetFeatureArea(path).getResult();
        while ((feature = layer.GetNextFeature()) != null) {
            Row row = sheet.createRow(rowIndex++);

            for (int i = 0; i < fieldCount; i++) {
                String fieldValue = feature.GetFieldAsString(i);
                row.createCell(i).setCellValue(fieldValue);
            }
            row.createCell(fieldCount).setCellValue(resultArea.get(feature.GetFieldAsString("id")));
            feature.delete();
        }
        var row = sheet.createRow(rowIndex++);
        row.createCell(0).setCellValue("总面积(平方公里)");
        row.createCell(fieldCount).setCellValue(resultArea.get("total"));
        sheet.addMergedRegion(new CellRangeAddress(rowIndex, rowIndex, 0, fieldCount));
        sheet.setAutoFilter(new CellRangeAddress(0, rowIndex, 0, fieldCount));
        // 写入到 Excel 文件
        try (FileOutputStream fos = new FileOutputStream(pathToSave)) {
            workbook.write(fos);
            System.out.println("表格导出成功: " + pathToSave);
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            workbook.close();
            dataset.delete();
        }
    }

    /***
     * 投影转换
     * @param path
     * @param epsgCode
     * @return
     */
    public static GdalResult<Void> TransFormProjection(String path,int epsgCode){
        var dataset = ogr.Open(path, 1); // Open in update mode
        if (dataset == null) {
            return new GdalResult<>("无法打开数据源", null);
        }
        var layer = dataset.GetLayer(0);
        if (layer == null) {
            dataset.delete();
            return new GdalResult<>("无法打开图层", null);
        }
        SpatialReference targetSRS = new SpatialReference();
        targetSRS.ImportFromEPSG(epsgCode);
        Feature feature;
        while ((feature = layer.GetNextFeature()) != null) {
            var geometry = feature.GetGeometryRef();
            if (geometry != null) {
                geometry.TransformTo(targetSRS);
            }
            feature.delete();
        }
        dataset.delete();
        return new GdalResult<>("转换完成",null);
    }

    /***
     * 联合分析
     * @param path
     * @return
     */
    public static GdalResult<String> UnionAnalysis(String path, String unionPath) {
        var dataset1 = ogr.Open(path, 0);
        var dataset2 = ogr.Open(unionPath, 0);

        if (dataset1 == null || dataset2 == null) {
            return new GdalResult<>("无法打开数据源", null);
        }

        var layer1 = dataset1.GetLayer(0);
        var layer2 = dataset2.GetLayer(0);

        if (layer1 == null || layer2 == null) {
            dataset1.delete();
            dataset2.delete();
            return new GdalResult<>("无法打开图层", null);
        }

        // 创建输出文件
        var driver = ogr.GetDriverByName("ESRI Shapefile");
        String savePath = path.split("\\.")[0] + "_union.shp";
        var file = new File(savePath);
        if (file.exists()) file.delete();

        var outputSource = driver.CreateDataSource(savePath);
        var outputLayer = outputSource.CreateLayer("union_result", layer1.GetSpatialRef(), ogr.wkbPolygon);

        // 拷贝属性表字段
        var layerDefn1 = layer1.GetLayerDefn();
        var layerDefn2 = layer2.GetLayerDefn();
        for (int i = 0; i < layerDefn1.GetFieldCount(); i++) {
            var fieldDefn = layerDefn1.GetFieldDefn(i);
            outputLayer.CreateField(fieldDefn);
        }
        for (int i = 0; i < layerDefn2.GetFieldCount(); i++) {
            var fieldDefn = layerDefn2.GetFieldDefn(i);
            outputLayer.CreateField(fieldDefn);
        }

        // 遍历两个图层，进行联合分析
        Feature feature1;
        while ((feature1 = layer1.GetNextFeature()) != null) {
            Feature feature2;
            while ((feature2 = layer2.GetNextFeature()) != null) {
                var geom1 = feature1.GetGeometryRef();
                var geom2 = feature2.GetGeometryRef();

                if (geom1 != null && geom2 != null) {
                    var unionGeom = geom1.Union(geom2); // 获取联合后的几何
                    if (unionGeom != null && !unionGeom.IsEmpty()) {
                        var outputFeature = new Feature(outputLayer.GetLayerDefn());
                        outputFeature.SetGeometry(unionGeom);

                        // 合并属性表字段
                        for (int i = 0; i < feature1.GetFieldCount(); i++) {
                            outputFeature.SetField(i, feature1.GetFieldAsString(i));
                        }
                        for (int i = 0; i < feature2.GetFieldCount(); i++) {
                            outputFeature.SetField(feature1.GetFieldCount() + i, feature2.GetFieldAsString(i));
                        }

                        outputLayer.CreateFeature(outputFeature);
                        outputFeature.delete();
                    }
                }
                feature2.delete();
            }
            feature1.delete();
            layer2.ResetReading(); // 重置 layer2 的读取状态
        }

        dataset1.delete();
        dataset2.delete();
        outputSource.delete();
        return new GdalResult<>("联合分析完成，结果保存至: " + savePath, savePath);
    }



    /***
     *裁剪
     * @param path
     * @param clipPath
     * @return
     */
    public static GdalResult<String> ClipAnalysis(String path,String clipPath){
        var dataset1 = ogr.Open(path, 0);
        var dataset2 = ogr.Open(clipPath, 0);
        if (dataset1 == null || dataset2 == null) {
            return new GdalResult<>("无法打开数据源", null);
        }
        var layer1 = dataset1.GetLayer(0);
        var layer2 = dataset2.GetLayer(0);

        if (layer1 == null || layer2 == null) {
            dataset1.delete();
            dataset2.delete();
            return new GdalResult<>("无法打开图层", null);
        }

        var driver = ogr.GetDriverByName("ESRI Shapefile");
        String savePath = path.split("\\.")[0] + "_clip.shp";
        var file = new File(savePath);
        if (file.exists()) file.delete();

        var outputSource = driver.CreateDataSource(savePath);
        var outputLayer = outputSource.CreateLayer("clip_result", layer1.GetSpatialRef(), ogr.wkbPolygon);

        Feature feature1;
        while ((feature1 = layer1.GetNextFeature()) != null) {
            Feature feature2;
            while ((feature2 = layer2.GetNextFeature()) != null) {
                var geom1 = feature1.GetGeometryRef();
                var geom2 = feature2.GetGeometryRef();

                if (geom1 != null && geom2 != null) {
                    var clippedGeom = geom1.Intersection(geom2);
                    if (clippedGeom != null && !clippedGeom.IsEmpty()) {
                        var outputFeature = new Feature(outputLayer.GetLayerDefn());
                        outputFeature.SetGeometry(clippedGeom);
                        outputLayer.CreateFeature(outputFeature);
                        outputFeature.delete();
                    }
                }
                feature2.delete();
            }
            feature1.delete();
        }

        dataset1.delete();
        dataset2.delete();
        outputSource.delete();
        return new GdalResult<>("裁剪分析完成 " ,null);
    }

    /***
     * 相交f
     * @param path
     * @param intersectionPath
     * @return
     */
    public static GdalResult<String> IntersectionAnalysis(String path, String intersectionPath) {
        return ClipAnalysis(path, intersectionPath); // 裁剪逻辑相同，可直接复用
    }


    /***
     * 交集取反
     * @param path
     * @param unIntPath
     * @return
     */
    public static GdalResult<String> UnIntersectAnalysis(String path,String unIntPath){
            var dataset1 = ogr.Open(path, 0);
            var dataset2 = ogr.Open(unIntPath, 0);

            if (dataset1 == null || dataset2 == null) {
                return new GdalResult<>("无法打开数据源", null);
            }

            var layer1 = dataset1.GetLayer(0);
            var layer2 = dataset2.GetLayer(0);

            if (layer1 == null || layer2 == null) {
                dataset1.delete();
                dataset2.delete();
                return new GdalResult<>("无法打开图层", null);
            }

            var driver = ogr.GetDriverByName("ESRI Shapefile");
            String savePath = path.split("\\.")[0] + "_unintersect.shp";
            var file = new File(savePath);
            if (file.exists()) file.delete();

            var outputSource = driver.CreateDataSource(savePath);
            var outputLayer = outputSource.CreateLayer("unintersect_result", layer1.GetSpatialRef(), ogr.wkbPolygon);

            Feature feature1;
            while ((feature1 = layer1.GetNextFeature()) != null) {
                Feature feature2;
                while ((feature2 = layer2.GetNextFeature()) != null) {
                    var geom1 = feature1.GetGeometryRef();
                    var geom2 = feature2.GetGeometryRef();

                    if (geom1 != null && geom2 != null) {
                        var diffGeom = geom1.SymmetricDifference(geom2);
                        if (diffGeom != null && !diffGeom.IsEmpty()) {
                            var outputFeature = new Feature(outputLayer.GetLayerDefn());
                            outputFeature.SetGeometry(diffGeom);
                            outputLayer.CreateFeature(outputFeature);
                            outputFeature.delete();
                        }
                    }
                    feature2.delete();
                }
                feature1.delete();
            }

            dataset1.delete();
            dataset2.delete();
            outputSource.delete();
            return new GdalResult<>("交集取反分析完成，结果保存至: " + savePath, savePath);
        }
    /***
     * 缓冲区分析
     * @param path
     * @param distance
     * @param quadsecs
     * @return
     */
    public static GdalResult<String> BufferAnalysis(String path,  double distance,int quadsecs){
       if (!OpenDataset(path)){
           return new GdalResult<>("无法打开数据源",null);
       }else{
           if (!OpenLayer(path)){
               return new GdalResult<>("无法打开图层",null);
           }
           var dataset = ogr.Open(path, 0);
           var layer = dataset.GetLayer(0);
           var driver = ogr.GetDriverByName("ESRI Shapefile");
           var savePath = path.split("\\.")[0]+"_buffer.shp";
           var file  = new File(savePath);
           if (file.exists()) {
               file.delete();
           }else{
               var outputSource = driver.CreateDataSource(savePath);
               var outputLayer =outputSource.CreateLayer(layer.GetName(),layer.GetSpatialRef(),layer.GetGeomType());
               for (int i = 0; i < layer.GetLayerDefn().GetFieldCount(); i++) {
                   var fieldDefn = layer.GetLayerDefn().GetFieldDefn(i);
                   outputLayer.CreateField(fieldDefn,i);
               }
               Feature feature;
               while ((feature = layer.GetNextFeature()) != null) {
                   var fieldCount = feature.GetFieldCount();
                   var geometry = feature.GetGeometryRef();
                   if (geometry != null) {
                       var buffer = geometry.Buffer(distance,quadsecs);

                       var newFeature = new Feature(outputLayer.GetLayerDefn());
                       newFeature.SetGeometry(buffer);
                       for (int i = 0; i < fieldCount; i++) {
                            newFeature.SetField(i,feature.GetFieldAsString(i));
                       }
                       outputLayer.CreateFeature(newFeature);
                       newFeature.delete();
                   }
               }
           }
           dataset.delete();
           layer.delete();
           return  new GdalResult<>("生成缓冲区成功",savePath);
       }
    }
    public static boolean OpenDataset(String path){
        var dataset = ogr.Open(path, 0);
        return dataset != null;
    }
    public static boolean OpenLayer(String path){
        var dataset = ogr.Open(path, 0);
        var layer = dataset.GetLayer(0);
        return layer != null;
    }
    public static void main(String[] args) throws IOException {
        InitGdal();
//        System.err.println(GetFeatureCount("E:\\data\\data.shp").getResult());
//        System.err.println(GetSummaryArea("E:\\data\\data.shp").getResult());
//        ExportAreaAsTable("E:\\data\\data.shp");
//       System.err.println( GetFeatureArea("E:\\data\\data.shp").getResult());
//        BufferAnalysis("E:\\data\\data.shp",1,36);
        UnionAnalysis("C:\\Users\\Administrator\\Documents\\a.shp","C:\\Users\\Administrator\\Documents\\b.shp");
    }

}
